#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Creating the training, validation, and test datasets

@author: vincent
"""

import os
import sys
import copy
import csv
import numpy as np
import time
import netCDF4 as nc
from scipy.interpolate import griddata

import torch
import torch.nn as nn

from funcsinputsNN import arrayStandardize
from funcsinputsNN import pastBoundedPeriodMean3dTensor


saving = True

### Directories ###
curdir      = os.getcwd()
DirRes      = f'{curdir}/OutputGlaDS/'
DirOut      = f'{curdir}/InputANN/'
TodayDate   = '05May2023'

### Set device ###
device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')

### Training glaciers used ###
lsTrGlaciers       = ['Jkb','Ptm','Hlm','Kgl','Hmb','Kgb','Rsl']

### Distribution of meltwater between moulins ###
runoffMoulinDistri = 'Uniform'

### Subdomain lower-left coordinates per glacier ###
subL       = 100e3 #length of sub-squares
dicSubDoms = {}
dicSubDoms['Jkb'] = [[-2321e3,-198e3],[-2321e3,-98e3]]
dicSubDoms['Ptm'] = [[-1125e3,-289e3],[-1025e3,-289e3]]
dicSubDoms['Hlm'] = [[-2590e3,217.8e3]]
dicSubDoms['Kgl'] = [[-2301e3,399.5e3]]
dicSubDoms['Hmb'] = [[-1178e3,194e3]]
dicSubDoms['Kgb'] = [[-2760e3,97.27e3]]
dicSubDoms['Rsl'] = [[-2566e3,-236e3]]
dicSubDoms['Upv'] = [[-1882e3-303e3]]

### Periods over which we calculate cumulative runoff ###
lsPersRnf   = [[-10/365.2524,0],[-1/12,0],[-2/12,-1/12],[-3/12,-2/12],[-4/12,-3/12],[-5/12,-4/12],[-6/12,-5/12],[-7/12,-6/12],[-8/12,-7/12],[-1,0]]
lsNamesPers = ['10days','month0tomonth1','month1tomonth2','month2tomonth3','month3tomonth4','month4tomonth5','month5tomonth6','month6tomonth7','month7tomonth8','1year']
nPersRnf    = len(lsPersRnf)

### Criteria to exclude irrelevant vertices ###
iceHmin    = 20.0 #[m] 
iceVelmin  = 5.0 #[m/yr]
phiminval  = 0 #[Pa] (any gridpoint with phi<phiminval at any timestep will be excluded)

### Options ###
StartTime       = 5 #time from which we keep the data
TimeTrEnd       = 31.0
Normalization   = True
nnx             = int(256/2)
nny             = int(256/2)
methodinterp    = 'linear'

### Initialize lists ###
lsgridx             = []
lsgridy             = []
lsbed               = []
lsbigh              = []
lsicevel            = []
lsxx1d              = []
lsyy1d              = []
lsisspc             = []
lsmoulinflux        = [] 
lsmoulinfluxprd0    = []
lsmoulinfluxprd1    = []
lsmoulinfluxprd2    = []
lsmoulinfluxprd3    = []
lsmoulinfluxprd4    = []
lsmoulinfluxprd5    = []
lsmoulinfluxprd6    = []
lsmoulinfluxprd7    = []
lsmoulinfluxprd8    = []
lsmoulinfluxprd9    = []
lsmonthmoulinflux   = []
lsday10moulinflux   = []
lsphi               = []
for gg,glname in enumerate(lsTrGlaciers):
    
    if(glname=='Jkb'):
        fileres = 'salJkbpc.nc'
    if(glname=='Hlm'):
        fileres = 'salHlmpc.nc'
    if(glname=='Ptm'):
        fileres = 'salPtmpc.nc'
    if(glname=='Kgl'):
        fileres = 'salKglpc.nc'
    if(glname=='Kgb'):
        fileres = 'salKgbpc.nc'
    if(glname=='Rsl'):
        fileres = 'salRslpc.nc'
    if(glname=='Hmb'):
        fileres = 'salHmbpc.nc'
    if(glname=='Upv'):
        fileres = 'salUpvpc.nc'

    ### Load data ###
    ds           = nc.Dataset(DirRes+fileres)
    timing00     = np.array(ds.variables['time'])
    isspc        = np.array(ds.variables['isSPC']).astype(int)
    bed          = np.array(ds.variables['Bed'])
    bighfull     = np.array(ds.variables['IceThickness'])
    icevelfull   = np.array(ds.variables['IceVel'])
    xx1d         = np.array(ds.variables['x'])
    yy1d         = np.array(ds.variables['y'])
    influxmoul00 = np.array(ds.variables['InfluxMoulins'])
    presmoul     = np.array(ds.variables['PresenceMoulin'])
    phifull0     = np.array(ds.variables['HydraulicPotential'])
    ds.close()
    
    ### Exclude time steps ###
    indstm            = np.logical_and(timing00>=StartTime,timing00<TimeTrEnd)
    indstm01          = timing00<TimeTrEnd
    timing0           = timing00[indstm01]
    indstm11          = timing0>=StartTime
    timing            = timing00[indstm]
    ntot              = len(timing)
    ntot00            = len(timing00)
    ntot0             = len(timing0) 
    phifull           = phifull0[:,indstm]
    influxmoul0       = influxmoul00[indstm01] #not removing pre-StartTime yet 

    if(runoffMoulinDistri=='Uniform'):
        nMouls      = np.sum(presmoul)
        moulinflux0 = np.array([influxmoul0[tt]*presmoul/nMouls for tt in range(ntot0)]).T

    ### Partition domain in 100x100 km2 pieces ###
    nsub = len(dicSubDoms[glname])
    for kk in range(nsub):
        lly = dicSubDoms[glname][kk][0]
        llx = dicSubDoms[glname][kk][1]
        # Grids #
        grx0 = np.zeros((nny,nnx))
        gry0 = np.zeros((nny,nnx))
        for ii in range(nny):
            grx0[ii,:] = np.linspace(llx,llx+subL,nnx)
        for jj in range(nnx):
            gry0[:,jj] = np.linspace(lly,lly+subL,nnx)
        lsgridx.append(grx0)
        lsgridy.append(gry0)
        ### Interpolate and save ###
        temp = griddata(np.vstack((yy1d,xx1d)).T,isspc,(gry0,grx0),methodinterp)
        lsisspc.append(temp)
        temp = griddata(np.vstack((yy1d,xx1d)).T,bed,(gry0,grx0),methodinterp)
        lsbed.append(temp)
        temp = griddata(np.vstack((yy1d,xx1d)).T,bighfull,(gry0,grx0),methodinterp)
        lsbigh.append(temp)
        temp = griddata(np.vstack((yy1d,xx1d)).T,icevelfull,(gry0,grx0),methodinterp)
        lsicevel.append(temp)
        gphi = np.zeros((ntot,nny,nnx))
        gfl0 = np.zeros((ntot0,nny,nnx))
        print(f'{glname}: gridding phi and moulin fluxes for subdomain {kk}/{nsub-1}')
        for tt in range(ntot):
            gphi[tt,:,:] = griddata(np.vstack((yy1d,xx1d)).T,phifull[:,tt],(gry0,grx0),methodinterp)
        if(np.any(np.isnan(gphi))):
            ValueError('nan in gridded values as grid extends beyond domain coordinates')
        lsphi.append(gphi)
        for tt in range(ntot0):
            gfl0[tt,:,:] = griddata(np.vstack((yy1d,xx1d)).T,moulinflux0[:,tt],(gry0,grx0),methodinterp)
        # Calculate cumulative runoff over past periods #
        if(nPersRnf>=1):
            gflprd = pastBoundedPeriodMean3dTensor(timing0,lsPersRnf[0][0],lsPersRnf[0][1],gfl0,axistime=0)
            lsmoulinfluxprd0.append(gflprd[indstm11,:,:])
        if(nPersRnf>=2):
            gflprd = pastBoundedPeriodMean3dTensor(timing0,lsPersRnf[1][0],lsPersRnf[1][1],gfl0,axistime=0)
            lsmoulinfluxprd1.append(gflprd[indstm11,:,:])
        if(nPersRnf>=3):
            gflprd = pastBoundedPeriodMean3dTensor(timing0,lsPersRnf[2][0],lsPersRnf[2][1],gfl0,axistime=0)
            lsmoulinfluxprd2.append(gflprd[indstm11,:,:])
        if(nPersRnf>=4):
            gflprd = pastBoundedPeriodMean3dTensor(timing0,lsPersRnf[3][0],lsPersRnf[3][1],gfl0,axistime=0)
            lsmoulinfluxprd3.append(gflprd[indstm11,:,:])
        if(nPersRnf>=5):
            gflprd = pastBoundedPeriodMean3dTensor(timing0,lsPersRnf[4][0],lsPersRnf[4][1],gfl0,axistime=0)
            lsmoulinfluxprd4.append(gflprd[indstm11,:,:])
        if(nPersRnf>=6):
            gflprd = pastBoundedPeriodMean3dTensor(timing0,lsPersRnf[5][0],lsPersRnf[5][1],gfl0,axistime=0)
            lsmoulinfluxprd5.append(gflprd[indstm11,:,:])
        if(nPersRnf>=7):
            gflprd = pastBoundedPeriodMean3dTensor(timing0,lsPersRnf[6][0],lsPersRnf[6][1],gfl0,axistime=0)
            lsmoulinfluxprd6.append(gflprd[indstm11,:,:])
        if(nPersRnf>=8):
            gflprd = pastBoundedPeriodMean3dTensor(timing0,lsPersRnf[7][0],lsPersRnf[7][1],gfl0,axistime=0)
            lsmoulinfluxprd7.append(gflprd[indstm11,:,:])
        if(nPersRnf>=9):
            gflprd = pastBoundedPeriodMean3dTensor(timing0,lsPersRnf[8][0],lsPersRnf[8][1],gfl0,axistime=0)
            lsmoulinfluxprd8.append(gflprd[indstm11,:,:])
        if(nPersRnf>=10):
            gflprd = pastBoundedPeriodMean3dTensor(timing0,lsPersRnf[9][0],lsPersRnf[9][1],gfl0,axistime=0)
            lsmoulinfluxprd9.append(gflprd[indstm11,:,:])

        lsmoulinflux.append(gfl0[indstm11,:,:])
        print(f'Done with {glname} subdomain {kk}/{nsub-1}')

### Total number of subdomains ###
ndom = len(lsbed)

### Get training data only for Normalization values ###
tmtrain      = timing<TimeTrEnd
nstepstr     = int(np.sum(tmtrain))
numtr        = int(nstepstr*ndom)

### Prepare data ###
nchin  = 4+nPersRnf #number of input channels
nchout = 1 #number of output channels
trdatainput  = np.zeros((numtr,nchin,nny,nnx))
trphitarget  = np.zeros((numtr,nchout,nny,nnx))
trtoexcl     = np.zeros((numtr,nchout,nny,nnx)).astype(bool)
tridsub      = np.zeros(numtr).astype(int)
trtiming     = np.zeros(numtr)

### Fill in training data tensors ###
for kk in range(ndom):
    # Training tensors #
    iitr = kk*nstepstr
    ttspc    = np.repeat(lsisspc[kk][np.newaxis,:,:],axis=0,repeats=nstepstr)
    ttbed    = np.repeat(lsbed[kk][np.newaxis,:,:],axis=0,repeats=nstepstr)
    ttbigh   = np.repeat(lsbigh[kk][np.newaxis,:,:],axis=0,repeats=nstepstr)
    tticevel = np.repeat(lsicevel[kk][np.newaxis,:,:],axis=0,repeats=nstepstr)
    trdatainput[iitr:iitr+nstepstr,0:4,:,:] = np.stack((ttbed,ttbigh,tticevel,lsmoulinflux[kk][tmtrain,:,:]),axis=1)
    if(nPersRnf>=1):
        trdatainput[iitr:iitr+nstepstr,4,:,:] = lsmoulinfluxprd0[kk][tmtrain,:,:]
    if(nPersRnf>=2):
        trdatainput[iitr:iitr+nstepstr,5,:,:] = lsmoulinfluxprd1[kk][tmtrain,:,:]
    if(nPersRnf>=3):
        trdatainput[iitr:iitr+nstepstr,6,:,:] = lsmoulinfluxprd2[kk][tmtrain,:,:]
    if(nPersRnf>=4):
        trdatainput[iitr:iitr+nstepstr,7,:,:] = lsmoulinfluxprd3[kk][tmtrain,:,:]
    if(nPersRnf>=5):
        trdatainput[iitr:iitr+nstepstr,8,:,:] = lsmoulinfluxprd4[kk][tmtrain,:,:]
    if(nPersRnf>=6):
        trdatainput[iitr:iitr+nstepstr,9,:,:] = lsmoulinfluxprd5[kk][tmtrain,:,:]
    if(nPersRnf>=7):
        trdatainput[iitr:iitr+nstepstr,10,:,:] = lsmoulinfluxprd6[kk][tmtrain,:,:]
    if(nPersRnf>=8):
        trdatainput[iitr:iitr+nstepstr,11,:,:] = lsmoulinfluxprd7[kk][tmtrain,:,:]
    if(nPersRnf>=9):
        trdatainput[iitr:iitr+nstepstr,12,:,:] = lsmoulinfluxprd8[kk][tmtrain,:,:]
    if(nPersRnf>=10):
        trdatainput[iitr:iitr+nstepstr,13,:,:] = lsmoulinfluxprd9[kk][tmtrain,:,:]
    
    trphitarget[iitr:iitr+nstepstr,0,:,:] = lsphi[kk][tmtrain,:,:]
    # Excluded vertices #
    ttminphi = np.repeat(np.min(lsphi[kk],axis=0)[np.newaxis,:,:],axis=0,repeats=ntot)
    inds     = np.logical_or(np.logical_or(np.logical_or(ttbigh<=iceHmin,tticevel<=iceVelmin),ttspc>0),ttminphi<phiminval)
    inds     = np.repeat(inds[:,np.newaxis,:,:],axis=0,repeats=nchout)
    trtoexcl[iitr:iitr+nstepstr,:,:,:] = np.copy(inds)
    # Sub-domain ID #
    tridsub[iitr:iitr+nstepstr] = kk
    # Timing #
    trtiming[iitr:iitr+nstepstr] = timing[tmtrain]
    
# Free memory #
#del(phifull,phifull0,gfl0,gflprd,gphi,inds,moulinflux0)
del(phifull,phifull0,gphi,inds,moulinflux0)
del(lsicevel,lsbigh,lsmoulinflux,lsday10moulinflux,lsmonthmoulinflux,lsphi)
del(lsmoulinfluxprd0,lsmoulinfluxprd1,lsmoulinfluxprd2,lsmoulinfluxprd3,lsmoulinfluxprd4,lsmoulinfluxprd5,lsmoulinfluxprd6,lsmoulinfluxprd7,lsmoulinfluxprd8,lsmoulinfluxprd9)

### Normalization values ###
dicMnSdev = {}
tempdata                          = trphitarget[trtoexcl==False]
temp,tempmn,tempsdev              = arrayStandardize(tempdata,getMeanAndSdev=True)
dicMnSdev['phi']                  = [tempmn,tempsdev]
tempdata                          = trdatainput[:,0,:,:][trtoexcl[:,0,:,:]==False]
temp,tempmn,tempsdev              = arrayStandardize(tempdata,getMeanAndSdev=True)
dicMnSdev['bedtopo']              = [tempmn,tempsdev]
tempdata                          = trdatainput[:,1,:,:][trtoexcl[:,0,:,:]==False]
temp,tempmn,tempsdev              = arrayStandardize(tempdata,getMeanAndSdev=True)
dicMnSdev['icethickness']         = [tempmn,tempsdev]
tempdata                          = trdatainput[:,2,:,:][trtoexcl[:,0,:,:]==False]
temp,tempmn,tempsdev              = arrayStandardize(tempdata,getMeanAndSdev=True)
dicMnSdev['icevel']               = [tempmn,tempsdev]
tempdata                          = trdatainput[:,3,:,:][trtoexcl[:,0,:,:]==False]
temp,tempmn,tempsdev              = arrayStandardize(tempdata,getMeanAndSdev=True)
dicMnSdev['moulinflux']           = [tempmn,tempsdev]
# Deal with the integrated moulin fluxes #
if(nPersRnf>=1):
    tempdata                    = trdatainput[:,4,:,:][trtoexcl[:,0,:,:]==False]
    temp,tempmn,tempsdev        = arrayStandardize(tempdata,getMeanAndSdev=True)
    dicMnSdev[lsNamesPers[0]]   = [tempmn,tempsdev]
if(nPersRnf>=2):
    tempdata                    = trdatainput[:,5,:,:][trtoexcl[:,0,:,:]==False]
    temp,tempmn,tempsdev        = arrayStandardize(tempdata,getMeanAndSdev=True)
    dicMnSdev[lsNamesPers[1]]   = [tempmn,tempsdev]
if(nPersRnf>=3):
    tempdata                    = trdatainput[:,6,:,:][trtoexcl[:,0,:,:]==False]
    temp,tempmn,tempsdev        = arrayStandardize(tempdata,getMeanAndSdev=True)
    dicMnSdev[lsNamesPers[2]]   = [tempmn,tempsdev]
if(nPersRnf>=4):
    tempdata                    = trdatainput[:,7,:,:][trtoexcl[:,0,:,:]==False]
    temp,tempmn,tempsdev        = arrayStandardize(tempdata,getMeanAndSdev=True)
    dicMnSdev[lsNamesPers[3]]   = [tempmn,tempsdev]
if(nPersRnf>=5):
    tempdata                    = trdatainput[:,8,:,:][trtoexcl[:,0,:,:]==False]
    temp,tempmn,tempsdev        = arrayStandardize(tempdata,getMeanAndSdev=True)
    dicMnSdev[lsNamesPers[4]]   = [tempmn,tempsdev]
if(nPersRnf>=6):
    tempdata                    = trdatainput[:,9,:,:][trtoexcl[:,0,:,:]==False]
    temp,tempmn,tempsdev        = arrayStandardize(tempdata,getMeanAndSdev=True)
    dicMnSdev[lsNamesPers[5]]   = [tempmn,tempsdev]
if(nPersRnf>=7):
    tempdata                    = trdatainput[:,10,:,:][trtoexcl[:,0,:,:]==False]
    temp,tempmn,tempsdev        = arrayStandardize(tempdata,getMeanAndSdev=True)
    dicMnSdev[lsNamesPers[6]]   = [tempmn,tempsdev]
if(nPersRnf>=8):
    tempdata                    = trdatainput[:,11,:,:][trtoexcl[:,0,:,:]==False]
    temp,tempmn,tempsdev        = arrayStandardize(tempdata,getMeanAndSdev=True)
    dicMnSdev[lsNamesPers[7]]   = [tempmn,tempsdev]
if(nPersRnf>=9):
    tempdata                    = trdatainput[:,12,:,:][trtoexcl[:,0,:,:]==False]
    temp,tempmn,tempsdev        = arrayStandardize(tempdata,getMeanAndSdev=True)
    dicMnSdev[lsNamesPers[8]]   = [tempmn,tempsdev]
if(nPersRnf>=10):
    tempdata                    = trdatainput[:,13,:,:][trtoexcl[:,0,:,:]==False]
    temp,tempmn,tempsdev        = arrayStandardize(tempdata,getMeanAndSdev=True)
    dicMnSdev[lsNamesPers[9]]   = [tempmn,tempsdev]

### Save Normalization values ###
if(saving):
    namevars = ['phi','bedtopo','icethickness','icevel','moulinflux']+lsNamesPers
    pathout  = DirSave+f'normalizationValues_{TodayDate}.csv'
    with open(pathout,'w',newline='') as outputcsv:
        header = ['Variable','Mean','Sdev']
        writer = csv.writer(outputcsv,header)
        writer.writerow(header)
        for ii in range(len(namevars)):
            writer.writerow([namevars[ii],dicMnSdev[namevars[ii]][0],dicMnSdev[namevars[ii]][1]])

       

